// Copyright (C) 1999 Dominic Letourneau and Andre Charbonneau
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this file.  If not, write to the Free Software Foundation,
// 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/*
** Includes
*/
#include "FileParser.h"
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <vector>
#include <list>
#include "FileParserKeys.h"





/*
** Private methods
*/

//------------------------------------------------------------------

string::size_type FileParser::get_start_index(const string& substring, const string& target, const string::size_type starting_at) const
{
  list< vector<int> > string_brackets_positions = this->get_string_brackets_positions(target);
  string::size_type return_value;

  return_value = target.find(substring,starting_at);
  while(in_string(return_value,string_brackets_positions))
    {
      return_value = target.find(substring,return_value + 1);
    }

  return return_value;
}

//------------------------------------------------------------------

string FileParser::read_file(const string& filename) throw(FileParserException)
{
  string file_content = "";
  char ch;
  char buf2[2];

  /* Check argument */
  if(filename == "")
    {
      /* Error, empty filename... Throw an exception.*/
      throw(FileParserException(IO_EXCEPTION,"Empty filename."));
    }
  ifstream input_file(filename.c_str());
  if(!input_file)
    {
      /* Error, could not open input file. Throw an exception. */
      throw(FileParserException(IO_EXCEPTION,string("Could not open data-flow text file:") + filename));
    }
  while(input_file.get(ch))
    {
      sprintf(buf2,"%c",ch);
      file_content += string(buf2);
    }
  
  /* set the m_filename member variable */
  m_filename = filename;

  return file_content;
}

//------------------------------------------------------------------

string FileParser::preprocess(const string& s) const throw(FileParserException)
{
  string return_string;

  try
    {
      /* First lets remove the comments... */
      return_string = this->remove_comments(s);
      
      /* Now lets remove the whitespaces... */
      return_string = this->remove_white_space(return_string);

      /* Now lets process environment vars... */
      return_string = this->process_env_vars(return_string);
    }
  catch(FileParserException)
    {
      // rethrow...
      throw;
    }

  return return_string;
}

//------------------------------------------------------------------

string FileParser::process_env_vars(const string& s) const throw(FileParserException)
{
  string::size_type index1,index2 = 0;
  string new_string = "";
  index1 = s.find("$(",0);
  while(index1 != string::npos)
    {
      if(index1 != 0)
	{
	  if(s[index1-1] == '\\')
	    {
	      continue;
	    }
	}
      string::size_type p_index = s.find(")",index1);
      if(p_index == string::npos)
	{
	  throw(FileParserException(SYNTAX_EXCEPTION,"Error in environment variable definition syntax."));
	}
      
      // get environment variable's value
      char *env_var = getenv(s.substr(index1+2, p_index-index1-2).c_str());
      if(env_var == NULL)
	{
	  throw(FileParserException(SYNTAX_EXCEPTION,string("Environment variable does not exist:") + s.substr(index1+2, p_index-index1-2)));
	}
      
      new_string += s.substr(index2,index1-index2);
      new_string += string(env_var);
      index2 = p_index+1;
      index1 = s.find("$(",index1+1);
  
    }
  if(index2 < s.length())
    {
      new_string += s.substr(index2, s.length() - index2);
    }

  return new_string;
}

//------------------------------------------------------------------

string FileParser::remove_comments(const string& s_arg) const throw(FileParserException)
{
  string s = s_arg;
  string::size_type start_of_comment_index = 0;
  string::size_type end_of_comment_index = 0;
  //string::size_type index = 0;
  bool done = false;

  while(!done)
    {
      start_of_comment_index = s.find("/*",0);
      if(start_of_comment_index == string::npos)
	{
	  done = true;
	  break;
	}

      end_of_comment_index = s.find("*/",start_of_comment_index+1);
      if(end_of_comment_index == string::npos)
	{
	  /* Comment mismatch, missing end of comment... */
	  throw(FileParserException(SYNTAX_EXCEPTION,"End of comment missing."));
	}
	     
      /* We have the index of the start and end of the comment, we can now remove it from the text. */
      s.replace(start_of_comment_index,(end_of_comment_index-start_of_comment_index)+2,"");

    }

  return s;
}

//------------------------------------------------------------------

string FileParser::remove_white_space(const string& s) const throw(FileParserException)
{
  string::size_type index = 0;
  string return_string = "";
  list< vector<int> > brackets_positions;

  try
    {
      brackets_positions = get_string_brackets_positions(s);
    }
  catch(FileParserException)
    {
      // simply rethrow...
      throw;
    }


  for(string::const_iterator i = s.begin(); i != s.end(); i++)
    {
      if((isspace((*i)) != 0) && !in_string(index,brackets_positions))
	{
	  // Blank character and not inside a string, remove it...
	}
      else
	{
	  // Non-blank character, add it to the return string.
	  return_string += (*i);
	}
      index++;
    }

  return return_string;
}

//------------------------------------------------------------------

bool FileParser::in_string(string::size_type index, list< vector<int> >& pairs_list) const throw(FileParserException)
{
  bool return_value = false;

  if(index != string::npos)
    {
      for(list< vector<int> >::const_iterator i = pairs_list.begin(); !(i == pairs_list.end()); i++)
	{
	  if( i->size() != 2 )
	    {
	      throw(FileParserException(FATAL_ERROR, "Bracket pair vector is not of size 2."));
	    }

	  if(index >= (*i)[0] && index <= (*i)[1])
	    {
	      return_value = true;
	      break;
	    }
	}
    }

  return return_value;
}

//------------------------------------------------------------------

list< vector<int> > FileParser::get_string_brackets_positions(const string& s) const throw(FileParserException)
{
  int index = 0;
  int pos1 = 0;
  bool toggle = false;
  string one_char_string = "";
  list<int> temp_int_list;
  list< vector<int> > pairs_list;
  vector<int> pair_vector(2);

  /* First lets create a list of all the double quotes indexes that are not escaped. */
  pos1 = s.find("\"",index);
  while(pos1 != string::npos)
    {
      /* check to see if it is escaped */
      if(pos1 > 0)
	{
	  one_char_string = s.substr(pos1-1,1);
	  if(one_char_string != "\\")
	    {
	      temp_int_list.push_back(pos1);
	    }
	}
      else
	{
	  temp_int_list.push_back(pos1);
	}

      index = pos1+1;
      pos1 = s.find("\"",index);      
    }

  /* We have a list of the positions of the non-escaped double quotes, lets group them in pairs. */
  /* First, check if the lenght of the list is even */
  if((temp_int_list.size() % 2) == 0)
    {
      for(list<int>::const_iterator i = temp_int_list.begin(); !(i == temp_int_list.end()); i++)
	{
	  if(toggle == false)
	    {
	      pair_vector[0] = (*i);
	    }
	  else
	    {
	      pair_vector[1] = (*i);
	      pairs_list.push_back(pair_vector);
	    }
	  toggle = !toggle;
	}
    }
  else
    {
      /* The number of non-escaped double quotes is not even, this means that a syntax error occored. */
      throw(FileParserException(SYNTAX_EXCEPTION,"String delimeter mismatch."));
    }

  return pairs_list;
}

//------------------------------------------------------------------

bool FileParser::node_blocks_brackets_ok(const string& s) throw(FileParserException)
{
  bool return_value;
  list<string::size_type> lt_indexes;
  list<string::size_type> gt_indexes;
  //bool toggle = false;
  string::size_type index = 0;
  
  index = get_start_index("<",s,0);
  while(index != string::npos)
    {
      lt_indexes.push_back(index);
      index = get_start_index("<",s,index + 1);
    }
  
  index = get_start_index(">",s,0);
  while(index != string::npos)
    {
      gt_indexes.push_back(index);
      index = get_start_index(">",s,index + 1);
    }

  if(lt_indexes.size() != gt_indexes.size())
    {
      return_value = false;
    }
  else
    {
      return_value = true;
      string::size_type lt_index, gt_index;
      while(!lt_indexes.empty())
	{
	  lt_index = lt_indexes.front();
	  lt_indexes.pop_front();
	  gt_index = gt_indexes.front();
	  gt_indexes.pop_front();
	  if(lt_index >= gt_index)
	    {
	      return_value = false;
	      break;
	    }
	}
    }

  return return_value;
}

//------------------------------------------------------------------

param_info FileParser::fill_param_info(const string& right_side, const string& net_name, const string& node_block) const throw(FileParserException)
{
  string type_key = "";
  string value = "";
  string::size_type comma_index1,colon_index;
  param_info pi;
  int ton = 0;
  
  
  comma_index1 = get_start_index(",",right_side,0);
  if(comma_index1 == string::npos)
    {
      // syntax error, missing comma in rightside
      throw(FileParserException(SYNTAX_EXCEPTION,string("Missing comma in parameter node block: ") + node_block + string("\n\tIn network: ") + net_name));
    }

  colon_index = get_start_index(":", right_side, comma_index1+1);

  pi.param_name = right_side.substr(0,comma_index1);

  if(colon_index != string::npos)
    {
      type_key = right_side.substr(comma_index1+1,colon_index - comma_index1 - 1);
      value = right_side.substr(colon_index+1);
    }
  else
    {
      type_key = "";
      value = right_side.substr(comma_index1 + 1);
    }
  
  if((pi.param_name == "") || (value == ""))
    {
      // syntax error,
      throw(FileParserException(SYNTAX_EXCEPTION,string("Missing information in parameter node block: ") + node_block + string("\n\tIn network: ") + net_name));
    }
  
  ton = this->type_of_num_string(value);

  if((value == "\"\"") && (type_key == "" || type_key == STRING_KEY))
    {
      // Empty string special case...
      pi.param_type = STRING_PARAM;
      pi.param_value = string("");
    }
  else
    if(((type_key == "") || (type_key == STRING_KEY)) && 
       (value[0] == '"') && (value[value.length()-1] == '"'))
      {
	// non-empty string
	pi.param_type = STRING_PARAM;
	pi.param_value = value.substr(1,value.length()-2);
      }
    else
      if(value == "''" && (type_key == "" || type_key == CHAR_KEY))
	{
	  // empty char sprecial case...
	  pi.param_type = CHAR_PARAM;
	  pi.param_value = string("");
	}
      else
	if((type_key == "" || type_key == CHAR_KEY) && 
	   (value[0] == '\'' && value[value.length()-1] == '\''))
	  {
	    // non-empty char case
	    if(value.size() != 3)
	      {
		throw(FileParserException(SYNTAX_EXCEPTION,string("Invalid character value found: ") + right_side + string("in node block: ") + node_block + string("\n\tIn network: ") + net_name));
	      }
	    pi.param_type = CHAR_PARAM;
	    pi.param_value = value.substr(1,value.length()-2);
	  }
	else
	  if(((type_key == "") && (value == "true" || value == "false")) ||
	     ((value == "true" || value == "false") && (type_key == BOOL_KEY)))
	    {
	      // boolean case with no type
	      pi.param_type = BOOL_PARAM;
	      pi.param_value = value;
	    }
	  else
	    if(type_key == BOOL_KEY && (value == "true" || value == "false"))
	      {
		throw(FileParserException(SYNTAX_EXCEPTION,string("Error in boolean value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
	      }
	    else
	      if(type_key == INT_KEY)
		{
		  if(!(ton == 1 || ton == 2))
		    {
		      throw(FileParserException(SYNTAX_EXCEPTION,string("Error in integer value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
		    }
		  pi.param_type = INT_PARAM;
		  pi.param_value = value;
		}
	      else
		if(type_key == SHORT_KEY)
		  {
		    if(!(ton == 1 || ton == 2))
		      {
			throw(FileParserException(SYNTAX_EXCEPTION,string("Error in short integer value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
		      }
		    pi.param_type = SHORT_PARAM;
		    pi.param_value = value;
		  }
		else
		  if(type_key == FLOAT_KEY)
		    {
		      if(!(ton == 3 || ton == 4))
			{
			  throw(FileParserException(SYNTAX_EXCEPTION,string("Error in float value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
			}
		      pi.param_type = FLOAT_PARAM;
		      pi.param_value = value;
		    }
		  else
		    if(type_key == DOUBLE_KEY)
		      {
			if(!(ton == 3 || ton == 4))
			  {
			    throw(FileParserException(SYNTAX_EXCEPTION,string("Error in double value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
			  }
			pi.param_type = DOUBLE_PARAM;
			pi.param_value = value;
		      }
		    else
		      if(type_key == LONG_KEY)
			{
			  if(!(ton == 1 || ton == 2))
			    {
			      throw(FileParserException(SYNTAX_EXCEPTION,string("Error in long integer value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
			    }
			  pi.param_type = LONG_PARAM;
			  pi.param_value = value;
			}
		      else
			if(type_key == UNSIGNED_CHAR_KEY)
			  {
			    if(!(ton == 1))
			      {
				throw(FileParserException(SYNTAX_EXCEPTION,string("Error in unsigned char value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
			      }
			    pi.param_type = UCHAR_PARAM;
			    pi.param_value = value;
			  }
			else
			  if(type_key == UNSIGNED_INT_KEY)
			    {
			      if(!(ton == 1))
				{
				  throw(FileParserException(SYNTAX_EXCEPTION,string("Error in unsigned int value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
				}
			      pi.param_type = UINT_PARAM;
			      pi.param_value = value;
			    }
			  else
			    if(type_key == UNSIGNED_LONG_KEY)
			      {
				if(!(ton == 1))
				  {
				    throw(FileParserException(SYNTAX_EXCEPTION,string("Error in unsigned long value in node block: ") + node_block + string("\n\tIn network: ") + net_name));
				  }
				pi.param_type = ULONG_PARAM;
				pi.param_value = value;
			      }
			    else
			      if(type_key == ENVIRONMENT_VARIABLE_KEY)
				{
				  char *env_var;
				  env_var = getenv(value.c_str());
				  if(env_var == NULL)
				    {
				      throw(FileParserException(SYNTAX_EXCEPTION,string("Invalid environment variable: ") + value + string(" in node block: ") + node_block + string("\n\tIn network: ") + net_name));
				    }
				  else
				    {
				      pi.param_type = STRING_PARAM;
				      pi.param_value = string(env_var);
				    }
				}
			      else
				if(type_key == SUBNET_PARAMETER_KEY)
				  {
				    pi.param_type = SUBNET_PARAM;
				    pi.param_value = value;
				  }
				else
				  if(type_key == EXTERNAL_PARAMETER_KEY)
				    {
				      pi.param_type = EXT_PARAM;
				      pi.param_value = value;
				    }
				  else
				    {
				      // no type key...
				      if(ton == 1 || ton == 2)
					{
					  pi.param_type = INT_PARAM;
					  pi.param_value = value;
					}
				      else
					if(ton == 3 || ton == 4)
					  {
					    pi.param_type = FLOAT_PARAM;
					    pi.param_value = value;
					  }
				      
				    }
  
  return pi;
}

//------------------------------------------------------------------

int FileParser::type_of_num_string(const string& s) const
{
  int return_value = 0;
  bool numeric = true;
  bool is_signed = false;
  bool floating_point = false;
  char c;

  for(unsigned int i = 0; i < s.length(); i++)
    {
      c = s[i];
      if(c == '-' && i == 0)
	{
	  is_signed = true;
	}
      else
      if(c == '.' && floating_point != true)
	{
	  floating_point = true;
	}
      else
      if(isdigit(c))
	{
	  // do nothing...
	}
      else
	{
	  numeric = false;
	  break;
	}
    }
  
  if(!numeric)
    {
      return_value = 0;
    }
  else if(!is_signed && !floating_point)
    {
      return_value = 1;
    }
  else if(is_signed && !floating_point)
    {
      return_value = 2;
    }
  else if(!is_signed && floating_point)
    {
      return_value = 3;
    }
  else
    {
      return_value = 4;
    }

  return return_value;
}

//------------------------------------------------------------------

map<string,FP_network_info_container> FileParser::process(const string& file_content) throw(FileParserException)
{
  map<string,FP_network_info_container> return_value;
  string::size_type index1, index2, left_block_bracket_index, right_block_bracket_index, lt_index, gt_index;
  bool bDone = false;

  string net_name = "";
  string node_name = "";

  string netblock_key = string(NET_BLOCK_KEY) + string(":");
  string iterator_net_key = string(ITERATOR_NET_BLOCK_KEY) + string(":");
  string node_key = NODE_KEY;
  string nodetype_key = NODE_TYPE_KEY;
  string input_key = INPUT_KEY;
  string param_key = PARAMETER_KEY;
  string input_node_key = NET_INPUT_KEY;
  string output_node_key = NET_OUTPUT_KEY;
  string condition_node_key = NET_CONDITION_KEY;

  bool iterator_mode = false;

  // First check for the alternance of the < and > brackets.
  if(!node_blocks_brackets_ok(file_content))
    {
      // test failed, so we have a syntax error
      throw(FileParserException(SYNTAX_EXCEPTION,"<> alternance check failure."));
    }



  index1 = 0;
  while(!bDone)
    {
      // Find a net or iterator block substring.  To do this we must find the next 
      // occurence for each of the two keys (netblock_key and iterator_net_key).  The
      // one that will occur first (before the other) will be chosen as the next
      // network delemiter key.
      string::size_type temp_index1 = index1;
      string::size_type temp_index2 = index1;

      temp_index1 = get_start_index(netblock_key,file_content,temp_index1);
      temp_index2 = get_start_index(iterator_net_key,file_content,temp_index2);

      if(((temp_index1 != string::npos) && (temp_index2 == string::npos)) ||
	 ((temp_index1 < temp_index2) && (temp_index1 != string::npos) && (temp_index2 != string::npos)))
	{
	  iterator_mode = false;
	  index1 = temp_index1;
	}
      else
      if(((temp_index1 == string::npos) && (temp_index2 != string::npos)) ||
	 ((temp_index2 < temp_index1) && (temp_index1 != string::npos) && (temp_index2 != string::npos)))
	{
	  iterator_mode = true;
	  index1 = temp_index2;
	}
      else
	{
	  index1 = string::npos;
	}


      // We have found a network delemiter... lets continue...

      if(index1 != string::npos)
	{
	  bool bNetBlockDone = false;
	  // find positions of the network delimeter characters. (that are not inside strings...)
	  left_block_bracket_index = get_start_index(NETWORK_START_DELIMETER,file_content,index1);
	  right_block_bracket_index = get_start_index(NETWORK_END_DELIMETER,file_content,index1);


	  if((left_block_bracket_index == string::npos) || (right_block_bracket_index == string::npos))
	    {
	      // syntax error...
	      throw(FileParserException(SYNTAX_EXCEPTION,string("Network block bracket missing.")));
	    }
	  if(left_block_bracket_index > right_block_bracket_index)
	    {
	      // syntax error...
	      throw(FileParserException(SYNTAX_EXCEPTION, "Network block bracket mismatch."));
	    }

	  // Extract the network name of the block
	  if(iterator_mode)
	    {
	      net_name = file_content.substr(index1+iterator_net_key.size(),left_block_bracket_index - (index1+iterator_net_key.size()));
	    }
	  else
	    {
	      net_name = file_content.substr(index1+netblock_key.size(),left_block_bracket_index - (index1+netblock_key.size()));
	    }

	  if(net_name == "")
	    {
	      // syntax error, empty net block name.
	      throw(FileParserException(SYNTAX_EXCEPTION, "Empty network name found."));
	    }

	  return_value.insert(map<string,FP_network_info_container>::value_type(net_name,FP_network_info_container(net_name,iterator_mode)));

	  index1 = right_block_bracket_index + 1;
	  index2 = left_block_bracket_index + 1;

	  while(!bNetBlockDone)
	    {
	      // find the next < character that is not in a string and that is before the end of block bracket
	      lt_index = get_start_index("<",file_content,index2);
	      // find the next > character that is not in a string and that is before the end of block bracket
	      gt_index = get_start_index(">",file_content,index2);
	      
	      if(lt_index == string::npos && gt_index == string::npos)
		{
		  bNetBlockDone = true;
		}
	      else
		{
		  // Check for syntax error on < > brackets...
		  if((lt_index < right_block_bracket_index) && (gt_index > right_block_bracket_index))
		    {
		      // syntax error
		      throw(FileParserException(SYNTAX_EXCEPTION,string("'>' missing inside network: ") + net_name));
		    }
		  
		  if((lt_index < right_block_bracket_index) && (gt_index < right_block_bracket_index)
		     && (lt_index != string::npos) && (gt_index != string::npos))
		    {
		      // process the <...> block.
		      string left_side = "";
		      string right_side = "";
		      string::size_type colon_index;
		      string node_block = "";
		      
		      index2 = gt_index + 1;
		      
		      node_block = file_content.substr(lt_index+1,gt_index-lt_index - 1);
		      if(node_block == "")
			{
			  // syntax error, empty node block...
			  throw(FileParserException(SYNTAX_EXCEPTION,string("Empty node block found in network: ") + net_name));
			}
		      colon_index = node_block.find(":",0);
		      if(colon_index == string::npos)
			{
			  // syntax error, no ":" in node block.
			  throw(FileParserException(SYNTAX_EXCEPTION,string("No ':' found in node block: ") + node_block + string("\n\tIn network: ") + net_name));
			}
		      left_side = node_block.substr(0,colon_index);
		      right_side = node_block.substr(colon_index+1, node_block.size() - (colon_index+1));
		      
		      if(left_side == "")
			{
			  // syntax error, missing left side of the net block
			  throw(FileParserException(SYNTAX_EXCEPTION,string("Left side missing in node block: ") + node_block + string("\n\tIn network: ") + net_name));
			}
		      if(right_side == "")
			{
			  // syntax error, missing right side of the net block
			  throw(FileParserException(SYNTAX_EXCEPTION,string("Right side missing in node block: ") + node_block + string("\n\tIn network: ") + net_name));
			}
		      
		      // process node block according to leftside (key)
		      if(left_side == node_key)
			{
			  map<string,FP_network_info_container>::iterator iter;
			  node_name = right_side;
			  iter = return_value.find(net_name);
			  if(iter != return_value.end())
			    {
			      iter->second.add_node(node_name);
			      if(verbose())
				{
				  cout << "Adding " << node_name << " to " << net_name << endl;
				}
			    }
			}
		      else
			if(left_side == nodetype_key)
			  {
			    map<string,FP_network_info_container>::iterator iter;
			    iter = return_value.find(net_name);
			    if(iter != return_value.end())
			      {
				iter->second.set_type_of_node(node_name,right_side);
			      }
			  }
			else
			  if(left_side == input_key)
			    {
			      map<string,FP_network_info_container>::iterator iter;
			      string::size_type comma_index1, comma_index2;
			      input_var_info ii;

			      comma_index1 = right_side.find(",",0);
			      if(comma_index1 == string::npos)
				{
				  // syntax error, missing comma in rightside
				  throw(FileParserException(SYNTAX_EXCEPTION,string("Missing 2 commas in input node block: ") + node_block + string("\n\tIn network: ") + net_name));
				}
			      comma_index2 = right_side.find(",",comma_index1 + 1);
			      if(comma_index2 == string::npos)
				{
				  // syntax error, missing comma in rightside
				  throw(FileParserException(SYNTAX_EXCEPTION,string("Missing 1 comma in input node block: ") + node_block + string("\n\tIn network: ") + net_name));
				}
			      if((comma_index2 - comma_index1) <= 1)
				{
				  // syntax error,
				  throw(FileParserException(SYNTAX_EXCEPTION,string("Missing information in input node block: ") + node_block + string("\n\tIn network: ") + net_name));
				}

			      ii.var_name = right_side.substr(0,comma_index1);
			      ii.source_node_name = right_side.substr(comma_index1+1,(comma_index2-comma_index1)-1);
			      ii.source_var_name = right_side.substr(comma_index2 + 1);

			      if((ii.var_name == "") || (ii.source_node_name == "") || (ii.source_var_name == ""))
				{
				  // syntax error,
				  throw(FileParserException(SYNTAX_EXCEPTION,string("Missing information in input node block: ") + node_block + string("\n\tIn network: ") + net_name));
				}

			      iter = return_value.find(net_name);
			      if(iter != return_value.end())
				{
				  iter->second.add_input_to_node(node_name,ii);
				}
			    }
			  else
			    if(left_side == param_key)
			      {
				map<string,FP_network_info_container>::iterator iter;
				param_info pi;
				try
				  {
				    pi = this->fill_param_info(right_side, net_name, node_block);
				  }
				catch(FileParserException)
				  {
				    throw;
				  }
				iter = return_value.find(net_name);
				if(iter != return_value.end())
				  {
				    iter->second.add_param_to_node(node_name,pi);
				  }
			      }
			    else 
			      if(left_side == input_node_key)
				{
				  map<string,FP_network_info_container>::iterator iter;
				  iter = return_value.find(net_name);
				  if(iter != return_value.end())
				    {
				      iter->second.set_input_node_name(right_side);
				    }
				}
			      else
				if(left_side == output_node_key)
				  {
				    map<string,FP_network_info_container>::iterator iter;
				    iter = return_value.find(net_name);
				    if(iter != return_value.end())
				      {
					iter->second.set_output_node_name(right_side);
				      }
				  }
				else
				  if(left_side == condition_node_key)
				    {
				      map<string,FP_network_info_container>::iterator iter;
				      iter = return_value.find(net_name);
				      if(iter != return_value.end())
					{
					  iter->second.set_condition_node_name(right_side);
					}
				    }
				  else
				    {
				      // syntax error, unkown node block key...
				      throw(FileParserException(SYNTAX_EXCEPTION,"Unknown node block key: " + left_side + string("\n\tIn network: ") + net_name));
				    }
		    }
		  else
		    {
		      bNetBlockDone = true;
		    }
		  
		}
	    }
	}
      else
	{
	  // no more network block in the RGB file, lets return...
	  bDone = true;
	}
    }

  return return_value;
}

//------------------------------------------------------------------

/*
** Public methods
*/

FileParser::FileParser()
{
  m_filename = "";
  m_verbose = false;
  return;
}

//------------------------------------------------------------------

FileParser::FileParser(const FileParser& fp)
{
  m_filename = fp.m_filename;
  m_verbose = fp.m_verbose;
  return;
}

//------------------------------------------------------------------

FileParser::~FileParser()
{
  // nothing do to...
  return;
}

//------------------------------------------------------------------

void FileParser::set_verbose(const bool b)
{
  m_verbose = b;
  return;
}

//------------------------------------------------------------------

bool FileParser::verbose() const
{
  return m_verbose;
}

//------------------------------------------------------------------

string FileParser::getFilename() const
{
  return m_filename;
}

//------------------------------------------------------------------

map<string,FP_network_info_container> FileParser::get_networks_map() const
{
  return m_networks_map;
}

//------------------------------------------------------------------

int FileParser::parse(const string& filename) throw(FileParserException)
{
  int return_value = 0;
  string file_content = "";
  
  m_filename = filename;
  
  if(verbose())
    {
      cout << "Parsing " << filename << endl;
    }

  try
    {
      // First lets load the RGB text file...
      file_content = this->read_file(filename);

      // Now lets preprocess the content of the file...
      file_content = this->preprocess(file_content);

      m_networks_map = process(file_content);

      return_value = m_networks_map.size();
    }
  catch(FileParserException)
    {
      throw;
    }
  catch(...)
    {
      cerr << "Unhandeled exception caught inside file parser. Re-throwing it..." << endl;
      throw;
    }

  if(verbose())
    {
      cout << "Parsing done." << endl;
    }

  return return_value;
}
